use std::io::{self, Write};
use std::char;
use std::str::FromStr;
use super::{ExpanderFunctions, expand_string};
use super::ranges::parse_index_range;

use types::Array;

// Bit Twiddling Guide:
// var & FLAG != 0 checks if FLAG is enabled
// var & FLAG == 0 checks if FLAG is disabled
// var |= FLAG enables the FLAG
// var &= 255 ^ FLAG disables the FLAG
// var ^= FLAG swaps the state of FLAG

const BACKSL: u8 = 1;
const SQUOTE: u8 = 2;
const DQUOTE: u8 = 4;
const EXPAND_PROCESSES: u8 = 8;

/// Index into a vector-like object
#[derive(Debug, PartialEq, Copy, Clone)]
pub enum Index {
    /// Index starting from the beginning of the vector, where `Forward(0)`
    /// is the first element
    Forward(usize),
    /// Index starting from the end of the vector, where `Backward(0)` is the
    /// last element. `
    Backward(usize)
}

impl Index {

    /// Construct an index using the following convetions:
    /// - A positive value `n` represents `Forward(n)`
    /// - A negative value `-n` reprents `Backwards(n - 1)` such that:
    /// ```
    /// assert_eq!(Index::new(-1), Index::Backward(0))
    /// ```
    pub fn new(input : isize) -> Index {
        if input < 0 {
            Index::Backward((input.abs() as usize) - 1)
        } else {
            Index::Forward(input.abs() as usize)
        }
    }

    pub fn resolve(&self, vector_length : usize) -> Option<usize> {
        match *self {
            Index::Forward(n) => Some(n),
            Index::Backward(n) => {
                if n >= vector_length {
                    None
                } else {
                   Some(vector_length - (n + 1))
                }
            }
        }
    }

}

/// A range of values in a vector-like object
#[derive(Debug, PartialEq, Copy, Clone)]
pub struct Range {
    /// Starting index
    start : Index,
    /// Ending index
    end : Index,
    /// Is this range inclusive? If false, this object represents a half-open
    /// range of [start, end), otherwise [start, end]
    inclusive: bool,
}

impl Range {

    pub fn to(end : Index) -> Range {
        Range {start : Index::new(0), end, inclusive : false}
    }

    pub fn from(start : Index) -> Range {
        Range {start , end : Index::new(-1), inclusive : true}
    }

    pub fn inclusive(start : Index, end : Index) -> Range {
        Range {start, end, inclusive : true}
    }

    pub fn exclusive(start : Index, end : Index) -> Range {
        Range {start, end, inclusive : false}
    }

    /// Returns the bounds of this range as a tuple containing:
    /// - The starting point of the range
    /// - The length of the range
    /// ```
    /// let vec = vec![0, 1, 2, 3, 4, 5, 6, 7, 8];
    /// let range = Range::exclusive(Index::new(1), Index::new(5));
    /// let (start, size) = range.bounds(vec.len()).unwrap();
    /// let expected = vec![1, 2, 3, 4];
    /// let selection = vec.iter().skip(start).take(size).collect::<Vec<_>>();
    /// assert_eq!(expected, selection);
    /// ```
    pub fn bounds(&self, vector_length : usize) -> Option<(usize,usize)> {
        if let Some(start) = self.start.resolve(vector_length) {
            if let Some(end) = self.end.resolve(vector_length) {
                if end < start {
                    None
                } else if self.inclusive {
                    Some((start, end - start + 1))
                } else {
                    Some((start, end - start))
                }
            } else {
                None
            }
        } else {
            None
        }
    }
}

/// Represents a filter on a vector-like object
#[derive(Debug, PartialEq, Copy, Clone)]
pub enum Select {
    /// Select no elements
    None,
    /// Select all elements
    All,
    /// Select a single element based on its index
    Index(Index),
    /// Select a range of elements
    Range(Range)
}

pub enum SelectError {
    Invalid
}

impl FromStr for Select {
    type Err = SelectError;
    fn from_str(data: &str) -> Result<Select, SelectError> {
        if ".." == data {
            return Ok(Select::All)
        }

        if let Ok(index) = data.parse::<isize>() {
            return Ok(Select::Index(Index::new(index)));
        }

        if let Some(range) = parse_index_range(data) {
            return Ok(Select::Range(range));
        }

        let stderr = io::stderr();
        let _ = writeln!(stderr.lock(), "ion: supplied selector, '{}', for array is invalid", data);

        Err(SelectError::Invalid)
    }
}

#[derive(Debug, PartialEq, Clone)]
enum Pattern<'a> {
    StringPattern(&'a str),
    Whitespace,
}

#[derive(Debug, PartialEq, Clone)]
pub struct ArrayMethod<'a> {
    method: &'a str,
    variable: &'a str,
    pattern: Pattern<'a>,
    selection: Select
}

impl<'a> ArrayMethod<'a> {
    pub fn handle(&self, current: &mut String, expand_func: &ExpanderFunctions) {
        match self.method {
            "split" => if let Some(variable) = (expand_func.variable)(self.variable, false) {
                match (&self.pattern, self.selection) {
                    (&Pattern::StringPattern(pattern), Select::All) => current.push_str (
                        &variable.split(&expand_string(pattern, expand_func, false).join(" "))
                            .collect::<Vec<&str>>()
                            .join(" ")
                    ),
                    (&Pattern::Whitespace, Select::All) => current.push_str (
                        &variable.split(char::is_whitespace)
                                 .collect::<Vec<&str>>()
                                 .join(" ")
                    ),
                    (_, Select::None) => (),
                    (&Pattern::StringPattern(pattern), Select::Index(Index::Forward(id))) => {
                        current.push_str(
                            variable.split(&expand_string(pattern, expand_func, false).join(" "))
                                    .nth(id)
                                    .unwrap_or_default())
                    } ,
                    (&Pattern::Whitespace, Select::Index(Index::Forward(id))) => current.push_str (
                        variable.split(char::is_whitespace)
                                .nth(id)
                                .unwrap_or_default()
                    ),
                    (&Pattern::StringPattern(pattern), Select::Index(Index::Backward(id))) => {
                        current.push_str(
                            variable.rsplit(&expand_string(pattern, expand_func, false).join(" "))
                                    .nth(id)
                                    .unwrap_or_default())
                    },
                    (&Pattern::Whitespace, Select::Index(Index::Backward(id))) => current.push_str(
                        variable.rsplit(char::is_whitespace)
                                .nth(id)
                                .unwrap_or_default()
                    ),
                    (&Pattern::StringPattern(pattern), Select::Range(range)) => {
                        let expansion = expand_string(pattern, expand_func, false).join(" ");
                        let iter = variable.split(&expansion);
                        if let Some((start, length)) = range.bounds(iter.clone().count()) {
                            let range = iter.skip(start)
                                            .take(length)
                                            .collect::<Vec<_>>()
                                            .join(" ");
                            current.push_str(&range)
                        }
                    },
                    (&Pattern::Whitespace, Select::Range(range)) => {
                        let len = variable.split(char::is_whitespace).count();
                        if let Some((start, length)) = range.bounds(len) {
                            let range = variable.split(char::is_whitespace)
                                                .skip(start)
                                                .take(length)
                                                .collect::<Vec<&str>>()
                                                .join(" ");
                            current.push_str(&range);
                        }

                    },
                }
            },
            _ => {
                let stderr = io::stderr();
                let mut stderr = stderr.lock();
                let _ = writeln!(stderr, "ion: invalid array method: {}", self.method);
            }
        }
    }

    pub fn handle_as_array(&self, expand_func: &ExpanderFunctions) -> Array {
        match self.method {
            "split" => if let Some(variable) = (expand_func.variable)(self.variable, false) {
                return match (&self.pattern, self.selection) {
                    (_, Select::None) => Some("".into()).into_iter().collect(),
                    (&Pattern::StringPattern(pattern), Select::All) => variable
                        .split(&expand_string(pattern, expand_func, false).join(" "))
                        .map(From::from)
                        .collect(),
                    (&Pattern::Whitespace, Select::All) => variable
                        .split(char::is_whitespace)
                        .map(From::from)
                        .collect(),
                    (&Pattern::StringPattern(pattern), Select::Index(Index::Forward(id))) =>
                            variable.split(&expand_string(pattern, expand_func, false).join(" "))
                                    .nth(id)
                                    .map(From::from)
                                    .into_iter()
                                    .collect(),
                    (&Pattern::Whitespace, Select::Index(Index::Forward(id))) =>
                            variable.split(char::is_whitespace)
                                    .nth(id)
                                    .map(From::from)
                                    .into_iter()
                                    .collect(),
                    (&Pattern::StringPattern(pattern), Select::Index(Index::Backward(id))) =>
                            variable.rsplit(&expand_string(pattern, expand_func, false).join(" "))
                                    .nth(id)
                                    .map(From::from)
                                    .into_iter()
                                    .collect(),
                    (&Pattern::Whitespace, Select::Index(Index::Backward(id))) =>
                            variable.rsplit(char::is_whitespace)
                                    .nth(id)
                                    .map(From::from)
                                    .into_iter()
                                    .collect(),
                    (&Pattern::StringPattern(pattern), Select::Range(range)) => {
                        let expansion = expand_string(pattern, expand_func, false).join(" ");
                        let iter = variable.split(&expansion);
                        if let Some((start, length)) = range.bounds(iter.clone().count()) {
                            iter.skip(start)
                                .take(length)
                                .map(From::from)
                                .collect()
                        } else {
                            Array::new()
                        }
                    },
                    (&Pattern::Whitespace, Select::Range(range)) => {
                        let len = variable.split(char::is_whitespace).count();
                        if let Some((start, length)) = range.bounds(len) {
                            variable.split(char::is_whitespace)
                                    .skip(start)
                                    .take(length)
                                    .map(From::from)
                                    .collect()
                        } else {
                            Array::new()
                        }
                    },
                }
            },
            _ => {
                let stderr = io::stderr();
                let mut stderr = stderr.lock();
                let _ = writeln!(stderr, "ion: invalid array method: {}", self.method);
            }
        }


        Some("".into()).into_iter().collect()
    }
}

#[derive(Debug, PartialEq, Clone)]
pub enum WordToken<'a> {
    Normal(&'a str,bool),
    Whitespace(&'a str),
    Tilde(&'a str),
    Brace(Vec<&'a str>),
    Array(Vec<&'a str>, Select),
    Variable(&'a str, bool, Select),
    ArrayVariable(&'a str, bool, Select),
    ArrayProcess(&'a str, bool, Select),
    Process(&'a str, bool, Select),
    StringMethod(&'a str, &'a str, &'a str, Select),
    ArrayMethod(ArrayMethod<'a>),
    //Glob(&'a str),
}

pub struct WordIterator<'a> {
    data:          &'a str,
    read:          usize,
    flags:         u8,
}

impl<'a> WordIterator<'a> {
    pub fn new(data: &'a str, expand_processes: bool) -> WordIterator<'a> {
        let flags = if expand_processes { EXPAND_PROCESSES } else { 0 };
        WordIterator { data: data, read: 0, flags: flags }
    }

    // Contains the grammar for collecting whitespace characters
    fn whitespaces<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let start = self.read;
        self.read += 1;
        while let Some(character) = iterator.next() {
            if character == b' ' {
                self.read += 1;
            } else {
                return WordToken::Whitespace(&self.data[start..self.read]);
            }
        }

        WordToken::Whitespace(&self.data[start..self.read])
    }

    /// Contains the logic for parsing tilde syntax
    fn tilde<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let start = self.read - 1;
        while let Some(character) = iterator.next() {
            match character {
                0...47 | 58...64 | 91...94 | 96 | 123...127 => {
                    return WordToken::Tilde(&self.data[start..self.read]);
                },
                _ => (),
            }
            self.read += 1;
        }

        WordToken::Tilde(&self.data[start..])
    }

    // Contains the logic for parsing braced variables
    fn braced_variable<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let start = self.read;
        while let Some(character) = iterator.next() {
            if character == b'}' {
                let output = &self.data[start..self.read];
                self.read += 1;
                return WordToken::Variable(output, self.flags & DQUOTE != 0, Select::All);
            }
            self.read += 1;
        }

        // The validator at the frontend should catch unterminated braced variables.
        panic!("ion: fatal error with syntax validation parsing: unterminated braced variable");
    }

    /// Contains the logic for parsing variable syntax
    fn variable<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let mut start = self.read;
        self.read += 1;
        while let Some(character) = iterator.next() {
            match character {
                b'(' => {
                    let method = &self.data[start..self.read];
                    self.read += 1;
                    start = self.read;
                    while let Some(character) = iterator.next() {
                        if character == b',' {
                            let variable = &self.data[start..self.read];
                            self.read += 1;
                            start = self.read;
                            while let Some(character) = iterator.next() {
                                if character == b')' {
                                    let pattern = &self.data[start..self.read].trim();
                                    self.read += 1;

                                    return if let Some(&b'[') = self.data.as_bytes().get(self.read) {
                                        let _ = iterator.next();
                                        WordToken::StringMethod(method, variable, pattern,
                                                                self.read_selection(iterator))
                                    } else {
                                        WordToken::StringMethod(method, variable, pattern, Select::All)
                                    };
                                }
                                self.read += 1;
                            }
                        } else if character == b')' {
                            // If no pattern is supplied, the default is a space.
                            let variable = &self.data[start..self.read];
                            self.read += 1;

                            return if let Some(&b'[') = self.data.as_bytes().get(self.read) {
                                let _ = iterator.next();
                                WordToken::StringMethod(method, variable, " ",
                                                        self.read_selection(iterator))
                            } else {
                                WordToken::StringMethod(method, variable, " ", Select::All)
                            };
                        }
                        self.read += 1;
                    }

                    panic!("ion: fatal error with syntax validation parsing: unterminated method");
                }
                // Only alphanumerical and underscores are allowed in variable names
                0...47 | 58...64 | 91...94 | 96 | 123...127 => {
                    let variable = &self.data[start..self.read];

                    return if character == b'[' {
                        WordToken::Variable(variable, self.flags & DQUOTE != 0,
                                            self.read_selection(iterator))
                    } else {
                        WordToken::Variable(variable, self.flags & DQUOTE != 0, Select::All)
                    };
                },
                _ => (),
            }
            self.read += 1;
        }

        WordToken::Variable(&self.data[start..], self.flags & DQUOTE != 0, Select::All)
    }

    fn read_selection<I>(&mut self, iterator: &mut I) -> Select
        where I: Iterator<Item = u8>
    {
        self.read += 1;
        let start = self.read;
        while let Some(character) = iterator.next() {
            if let b']' = character {
                let selection = match self.data[start..self.read].parse::<Select>() {
                    Ok(selection) => selection,
                    Err(_)       => Select::None
                };
                self.read += 1;
                return selection
            }
            self.read += 1;
        }

        panic!()
    }

    /// Contains the logic for parsing array variable syntax
    fn array_variable<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let mut start = self.read;
        self.read += 1;
        while let Some(character) = iterator.next() {
            match character {
                b'(' => {
                    let method = &self.data[start..self.read];
                    self.read += 1;
                    start = self.read;
                    while let Some(character) = iterator.next() {
                        if character == b',' {
                            let variable = &self.data[start..self.read];
                            self.read += 1;
                            start = self.read;
                            while let Some(character) = iterator.next() {
                                if character == b')' {
                                    let pattern = &self.data[start..self.read].trim();
                                    self.read += 1;
                                    return if let Some(&b'[') = self.data.as_bytes().get(self.read) {
                                        let _ = iterator.next();
                                        WordToken::ArrayMethod(ArrayMethod {
                                            method,
                                            variable,
                                            pattern: Pattern::StringPattern(pattern),
                                            selection: self.read_selection(iterator)
                                        })
                                    } else {
                                        WordToken::ArrayMethod(ArrayMethod {
                                            method,
                                            variable,
                                            pattern: Pattern::StringPattern(pattern),
                                            selection: Select::All
                                        })
                                    }
                                }
                                self.read += 1;
                            }
                        } else if character == b')' {
                            // If no pattern is supplied, the default is a space.
                            let variable = &self.data[start..self.read];
                            self.read += 1;

                            return if let Some(&b'[') = self.data.as_bytes().get(self.read) {
                                let _ = iterator.next();
                                WordToken::ArrayMethod(ArrayMethod {
                                    method,
                                    variable,
                                    pattern: Pattern::Whitespace,
                                    selection: self.read_selection(iterator)
                                })
                            } else {
                                WordToken::ArrayMethod(ArrayMethod {
                                    method,
                                    variable,
                                    pattern: Pattern::Whitespace,
                                    selection: Select::All
                                })
                            }
                        }
                        self.read += 1;
                    }

                    panic!("ion: fatal error with syntax validation parsing: unterminated method");
                },
                b'[' => {
                    return WordToken::ArrayVariable (
                        &self.data[start..self.read],
                        self.flags & DQUOTE != 0,
                        self.read_selection(iterator)
                    );
                },
                // Only alphanumerical and underscores are allowed in variable names
                0...47 | 58...64 | 91...94 | 96 | 123...127 => {
                    return WordToken::ArrayVariable(&self.data[start..self.read],
                                                    self.flags & DQUOTE != 0,
                                                    Select::All);
                },
                _ => (),
            }
            self.read += 1;
        }

        WordToken::ArrayVariable(&self.data[start..], self.flags & DQUOTE != 0, Select::All)
    }

    /// Contains the logic for parsing subshell syntax.
    fn process<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let start = self.read;
        let mut level = 0;
        while let Some(character) = iterator.next() {
            match character {
                _ if self.flags & BACKSL != 0     => self.flags ^= BACKSL,
                b'\\'                             => self.flags ^= BACKSL,
                b'\'' if self.flags & DQUOTE == 0 => self.flags ^= SQUOTE,
                b'"'  if self.flags & SQUOTE == 0 => self.flags ^= DQUOTE,
                b'$'  if self.flags & SQUOTE == 0 => {
                    if self.data.as_bytes()[self.read+1] == b'(' {
                        level += 1;
                    }
                },
                b')' if self.flags & SQUOTE == 0 => {
                    if level == 0 {
                        let output = &self.data[start..self.read];
                        self.read += 1;
                        return if let Some(&b'[') = self.data.as_bytes().get(self.read) {
                            let _ = iterator.next();
                            WordToken::Process(
                                output,
                                self.flags & DQUOTE != 0,
                                self.read_selection(iterator)
                            )
                        } else {
                            WordToken::Process(
                                output,
                                self.flags & DQUOTE != 0,
                                Select::All
                            )
                        }
                    } else {
                        level -= 1;
                    }
                }
                _ => (),
            }
            self.read += 1;
        }

        // The validator at the frontend should catch unterminated processes.
        panic!("ion: fatal error with syntax validation: unterminated process");
    }

    /// Contains the logic for parsing array subshell syntax.
    fn array_process<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let start = self.read;
        let mut level = 0;
        while let Some(character) = iterator.next() {
            match character {
                _ if self.flags & BACKSL != 0     => self.flags ^= BACKSL,
                b'\\'                             => self.flags ^= BACKSL,
                b'\'' if self.flags & DQUOTE == 0 => self.flags ^= SQUOTE,
                b'"'  if self.flags & SQUOTE == 0 => self.flags ^= DQUOTE,
                b'@'  if self.flags & SQUOTE == 0 => {
                    if self.data.as_bytes()[self.read+1] == b'[' {
                        level += 1;
                    }
                },
                b']' if self.flags & SQUOTE == 0 => {
                    if level == 0 {
                        let array_process_contents = &self.data[start..self.read];
                        self.read += 1;
                        return if let Some(&b'[') = self.data.as_bytes().get(self.read) {
                            let _ = iterator.next();
                            WordToken::ArrayProcess (
                                array_process_contents,
                                self.flags & DQUOTE != 0,
                                self.read_selection(iterator)
                            )
                        } else {
                            WordToken::ArrayProcess (
                                array_process_contents,
                                self.flags & DQUOTE != 0,
                                Select::All
                            )
                        }
                    } else {
                        level -= 1;
                    }
                }
                _ => (),
            }
            self.read += 1;
        }

        // The validator at the frontend should catch unterminated processes.
        panic!("ion: fatal error with syntax validation: unterminated array process");
    }

    /// Contains the grammar for parsing brace expansion syntax
    fn braces<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let mut start = self.read;
        let mut level = 0;
        let mut elements = Vec::new();
        while let Some(character) = iterator.next() {
            match character {
                _ if self.flags & BACKSL != 0     => self.flags ^= BACKSL,
                b'\\'                             => self.flags ^= BACKSL,
                b'\'' if self.flags & DQUOTE == 0 => self.flags ^= SQUOTE,
                b'"'  if self.flags & SQUOTE == 0 => self.flags ^= DQUOTE,
                b','  if self.flags & (SQUOTE + DQUOTE) == 0 && level == 0 => {
                    elements.push(&self.data[start..self.read]);
                    start = self.read + 1;
                },
                b'{' if self.flags & (SQUOTE + DQUOTE) == 0 => level += 1,
                b'}' if self.flags & (SQUOTE + DQUOTE) == 0 => {
                    if level == 0 {
                        elements.push(&self.data[start..self.read]);
                        self.read += 1;
                        return WordToken::Brace(elements);
                    } else {
                        level -= 1;
                    }

                },
                _ => ()
            }
            self.read += 1;
        }

        panic!("ion: fatal error with syntax validation: unterminated brace")
    }

    /// Contains the grammar for parsing array expression syntax
    fn array<I>(&mut self, iterator: &mut I) -> WordToken<'a>
        where I: Iterator<Item = u8>
    {
        let mut start = self.read;
        let mut level = 0;
        let mut whitespace = true;
        let mut elements = Vec::new();
        while let Some(character) = iterator.next() {
            match character {
                _ if self.flags & BACKSL != 0     => self.flags ^= BACKSL,
                b'\\'                             => self.flags ^= BACKSL,
                b'\'' if self.flags & DQUOTE == 0 => self.flags ^= SQUOTE,
                b'"'  if self.flags & SQUOTE == 0 => self.flags ^= DQUOTE,
                b' '  if self.flags & (SQUOTE + DQUOTE) == 0 && level == 0 => {
                    if whitespace {
                        self.read += 1;
                        start = self.read;
                    } else {
                        elements.push(&self.data[start..self.read]);
                        start = self.read + 1;
                        self.read += 1;
                        whitespace = true;
                    }
                    continue
                },
                b'[' if self.flags & (SQUOTE + DQUOTE) == 0 => level += 1,
                b']' if self.flags & (SQUOTE + DQUOTE) == 0 => {
                    if level == 0 {
                        elements.push(&self.data[start..self.read]);
                        self.read += 1;

                        return if let Some(&b'[') = self.data.as_bytes().get(self.read) {
                            let _ = iterator.next();
                            WordToken::Array(elements, self.read_selection(iterator))
                        } else {
                            WordToken::Array(elements, Select::All)

                        }
                    } else {
                        level -= 1;
                    }

                },
                _ => whitespace = false
            }
            self.read += 1;
        }

        panic!("ion: fatal error with syntax validation: unterminated array expression")
    }

    fn glob_check<I>(&mut self, iterator: &mut I) -> bool
        where I: Iterator<Item = u8> + Clone
    {
        // Clone the iterator and scan for illegal characters until the corresponding ] is
        // discovered. If none are found, then it's a valid glob signature.
        let mut moves = 0;
        let mut glob = false;
        let mut square_bracket = 0;
        let mut iter = iterator.clone();
        while let Some(character) = iter.next() {
            moves+=1;
            match character {
                b'[' => {
                    square_bracket += 1;
                }
                b' '| b'"' | b'\'' | b'$' | b'{' | b'}' => {
                    break;
                },
                b']' => {
                    // If the glob is less than three bytes in width, then it's empty and thus invalid.
                    if !(moves<=3 && square_bracket == 1) {
                        glob = true;
                        break
                    }
                }
                _=> (),
            }
        }

        if glob {
            for _ in 0..moves {
                iterator.next();
            }
            self.read += moves+1;
            true
        } else {
            self.read += 1;
            false
        }
    }
}


impl<'a> Iterator for WordIterator<'a> {
    type Item = WordToken<'a>;

    fn next(&mut self) -> Option<WordToken<'a>> {
        if self.read == self.data.len() { return None }

        let mut iterator = self.data.bytes().skip(self.read);
        let mut start = self.read;
        let mut glob = false;
        loop {

            if let Some(character) = iterator.next() {
                match character {
                    _ if self.flags & BACKSL != 0 => {
                        self.read += 1;
                        self.flags ^= BACKSL;
                        break
                    },
                    b'\\' => {
                        if self.flags & (DQUOTE + SQUOTE) == 0 { start += 1; }
                        self.read += 1;
                        self.flags ^= BACKSL;
                        if self.flags & EXPAND_PROCESSES == 0 {
                            return Some(WordToken::Normal("\\",glob));
                        }
                        break
                    }
                    b'\'' if self.flags & DQUOTE == 0 => {
                        start += 1;
                        self.read += 1;
                        self.flags ^= SQUOTE;
                        if self.flags & EXPAND_PROCESSES == 0 {
                            return Some(WordToken::Normal("'",glob));
                        }
                    },
                    b'"' if self.flags & SQUOTE == 0 => {
                        start += 1;
                        self.read += 1;
                        self.flags ^= DQUOTE;
                        if self.flags & EXPAND_PROCESSES == 0 {
                            return Some(WordToken::Normal("\"",glob));
                        }
                    }
                    b' ' if self.flags & (SQUOTE + DQUOTE) == 0 => {
                        return Some(self.whitespaces(&mut iterator));
                    }
                    b'~' if self.flags & (SQUOTE + DQUOTE) == 0 => {
                        self.read += 1;
                        return Some(self.tilde(&mut iterator));
                    },
                    b'{' if self.flags & (SQUOTE + DQUOTE) == 0 => {
                        self.read += 1;
                        return Some(self.braces(&mut iterator));
                    },
                    b'[' if self.flags & SQUOTE == 0 => {
                        if self.glob_check(&mut iterator) {
                            glob = true;
                        } else {
                            return Some(self.array(&mut iterator));
                        }
                    },
                    b'@' if self.flags & SQUOTE == 0 => {
                        match iterator.next() {
                            Some(b'[') => {
                                self.read += 2;
                                return if self.flags & EXPAND_PROCESSES != 0 {
                                    Some(self.array_process(&mut iterator))
                                } else {
                                    Some(WordToken::Normal(&self.data[start..self.read],glob))
                                }
                            },
                            // Some(b'{') => {
                            //     self.read += 2;
                            //     return Some(self.braced_variable(&mut iterator));
                            // }
                            _ => {
                                self.read += 1;
                                return Some(self.array_variable(&mut iterator));
                            }
                        }
                    }
                    b'$' if self.flags & SQUOTE == 0 => {
                        match iterator.next() {
                            Some(b'(') => {
                                self.read += 2;
                                return if self.flags & EXPAND_PROCESSES != 0 {
                                    Some(self.process(&mut iterator))
                                } else {
                                    Some(WordToken::Normal(&self.data[start..self.read],glob))
                                }
                            },
                            Some(b'{') => {
                                self.read += 2;
                                return Some(self.braced_variable(&mut iterator));
                            }
                            _ => {
                                self.read += 1;
                                return Some(self.variable(&mut iterator));
                            }
                        }
                    },
                    b'*'|b'?' => {
                        // if a word is not special, make sure you return the globbed variant at the end
                        self.read+=1;
                        glob=true; //warning is incorrect it does get read
                    },
                    _ => { self.read += 1; break },
                }
            } else {
                return None
            }
        }

        while let Some(character) = iterator.next() {
            match character {
                _ if self.flags & BACKSL != 0 => self.flags ^= BACKSL,
                b'\\' => {
                    self.flags ^= BACKSL;
                    let end = if self.flags & EXPAND_PROCESSES == 0 {
                        if self.flags & (DQUOTE + SQUOTE) != 0 { self.read+2 } else { self.read+1 }
                    } else if self.flags & (DQUOTE + SQUOTE) != 0 {
                        self.read+1
                    } else {
                        self.read
                    };
                    let output = &self.data[start..end];
                    self.read += 1;
                    return Some(WordToken::Normal(output,glob));
                },
                b'\'' if self.flags & DQUOTE == 0 => {
                    self.flags ^= SQUOTE;
                    let end = if self.flags & EXPAND_PROCESSES == 0 { self.read+1 } else { self.read };
                    let output = &self.data[start..end];
                    self.read += 1;
                    return Some(WordToken::Normal(output,glob));
                },
                b'"' if self.flags & SQUOTE == 0 => {
                    self.flags ^= DQUOTE;
                    let end = if self.flags & EXPAND_PROCESSES == 0 { self.read+1 } else { self.read };
                    let output = &self.data[start..end];
                    self.read += 1;
                    return Some(WordToken::Normal(output,glob));
                },
                b' ' | b'{' if self.flags & (SQUOTE + DQUOTE) == 0 => {
                    return Some(WordToken::Normal(&self.data[start..self.read],glob));
                },
                b'$' | b'@' if self.flags & SQUOTE == 0 => {
                    return Some(WordToken::Normal(&self.data[start..self.read],glob));
                },
                b'[' if self.flags & SQUOTE == 0 => {
                    if self.glob_check(&mut iterator) {
                        glob = true;
                    } else {
                        return Some(WordToken::Normal(&self.data[start..self.read],glob));
                    }
                },
                b'*'|b'?' if self.flags & SQUOTE == 0 => {
                    // if a word is not special, make sure you return the globbed variant at the end
                    self.read += 1;
                    glob = true; //warning is incorrect it does get read
                },
                _ => (),
            }
            self.read += 1;
        }

        if start == self.read {
            None
        } else {
            //println!("Normal exit");
            Some(WordToken::Normal(&self.data[start..],glob))
        }
    }
}


#[cfg(test)]
mod tests {
    use super::*;

    fn compare(input: &str, expected: Vec<WordToken>) {
        let mut correct = 0;
        for (actual, expected) in WordIterator::new(input, true).zip(expected.iter()) {
            assert_eq!(actual, *expected, "{:?} != {:?}", actual, expected);
            correct += 1;
        }
        assert_eq!(expected.len(), correct);
    }

    #[test]
    fn ranges() {
        let range1 = Range::exclusive(Index::new(1), Index::new(5));
        assert_eq!(Some((1, 4)), range1.bounds(42));
        assert_eq!(Some((1, 4)), range1.bounds(7));
        let range2 = Range::inclusive(Index::new(2), Index::new(-4));
        assert_eq!(Some((2, 5)), range2.bounds(10));
        assert_eq!(None, range2.bounds(3));
    }


    #[test]
    fn string_method() {
        let input = "$join(array, 'pattern') $join(array, 'pattern')";
        let expected = vec![
            WordToken::StringMethod("join", "array", "'pattern'", Select::All),
            WordToken::Whitespace(" "),
            WordToken::StringMethod("join", "array", "'pattern'", Select::All)
        ];
        compare(input, expected);
    }

    #[test]
    fn escape_with_backslash() {
        let input = "\\$FOO\\$BAR \\$FOO";
        let expected =
            vec![
                WordToken::Normal("$FOO",false),
                WordToken::Normal("$BAR",false),
                WordToken::Whitespace(" "),
                WordToken::Normal("$FOO",false)
            ];
        compare(input, expected);
    }

    #[test]
    fn array_expressions() {
        let input = "[ one two [three four]] [[one two] three four][0]";
        let first = vec![ "one", "two", "[three four]"];
        let second = vec![ "[one two]", "three", "four"];
        let expected = vec![
            WordToken::Array(first, Select::All),
            WordToken::Whitespace(" "),
            WordToken::Array(second, Select::Index(Index::new(0))),
        ];
        compare(input, expected);
    }

    #[test]
    fn array_variables() {
        let input = "@array @array[0]";
        let expected = vec![
            WordToken::ArrayVariable("array", false, Select::All),
            WordToken::Whitespace(" "),
            WordToken::ArrayVariable("array", false, Select::Index(Index::new(0))),
        ];
        compare(input, expected);
    }

    #[test]
    fn array_processes() {
        let input = "@[echo one two three] @[echo one two three][0]";
        let expected = vec![
            WordToken::ArrayProcess("echo one two three", false, Select::All),
            WordToken::Whitespace(" "),
            WordToken::ArrayProcess("echo one two three", false, Select::Index(Index::new(0)))
        ];
        compare(input, expected);
    }

    #[test]
    fn indexes() {
        let input = "@array[0..3] @array[0...3] @array[abc] @array[..3] @array[3..]";
        let expected = vec![
            WordToken::ArrayVariable("array", false, Select::Range(Range::exclusive(Index::new(0),
                                                                                    Index::new(3)))),
            WordToken::Whitespace(" "),
            WordToken::ArrayVariable("array", false, Select::Range(Range::inclusive(Index::new(0),
                                                                                    Index::new(3)))),
            WordToken::Whitespace(" "),
            WordToken::ArrayVariable("array", false, Select::None),
            WordToken::Whitespace(" "),
            WordToken::ArrayVariable("array", false, Select::Range(Range::to(Index::new(3)))),
            WordToken::Whitespace(" "),
            WordToken::ArrayVariable("array", false, Select::Range(Range::from(Index::new(3)))),
        ];
        compare(input, expected);
    }

    #[test]
    fn nested_processes() {
        let input = "echo $(echo $(echo one)) $(echo one $(echo two) three)";
        let expected = vec![
            WordToken::Normal("echo",false),
            WordToken::Whitespace(" "),
            WordToken::Process("echo $(echo one)", false, Select::All),
            WordToken::Whitespace(" "),
            WordToken::Process("echo one $(echo two) three", false, Select::All),
        ];
        compare(input, expected);
    }

    #[test]
    fn words_process_with_quotes() {
        let input = "echo $(git branch | rg '[*]' | awk '{print $2}')";
        let expected = vec![
            WordToken::Normal("echo",false),
            WordToken::Whitespace(" "),
            WordToken::Process("git branch | rg '[*]' | awk '{print $2}'", false, Select::All),
        ];
        compare(input, expected);

        let input = "echo $(git branch | rg \"[*]\" | awk '{print $2}')";
        let expected = vec![
            WordToken::Normal("echo",false),
            WordToken::Whitespace(" "),
            WordToken::Process("git branch | rg \"[*]\" | awk '{print $2}'", false, Select::All),
        ];
        compare(input, expected);
    }

    #[test]
    fn test_words() {
        let input = "echo $ABC \"${ABC}\" one{$ABC,$ABC} ~ $(echo foo) \"$(seq 1 100)\"";
        let expected = vec![
            WordToken::Normal("echo",false),
            WordToken::Whitespace(" "),
            WordToken::Variable("ABC", false, Select::All),
            WordToken::Whitespace(" "),
            WordToken::Variable("ABC", true, Select::All),
            WordToken::Whitespace(" "),
            WordToken::Normal("one",false),
            WordToken::Brace(vec!["$ABC", "$ABC"]),
            WordToken::Whitespace(" "),
            WordToken::Tilde("~"),
            WordToken::Whitespace(" "),
            WordToken::Process("echo foo", false, Select::All),
            WordToken::Whitespace(" "),
            WordToken::Process("seq 1 100", true, Select::All)
        ];
        compare(input, expected);
    }

    #[test]
    fn test_multiple_escapes() {
        let input = "foo\\(\\) bar\\(\\)";
        let expected = vec![
            WordToken::Normal("foo",false),
            WordToken::Normal("(",false),
            WordToken::Normal(")",false),
            WordToken::Whitespace(" "),
            WordToken::Normal("bar",false),
            WordToken::Normal("(",false),
            WordToken::Normal(")",false),
        ];
        compare(input, expected);
    }
}
